import { v4 as uuidv4 } from 'uuid';
import Base from './base';
import { onClickOutside } from '../ui_utils'
import { getAbsoluteCoordinate } from './position';
import { getClickPoint, createSvg } from './utils';
import { MARGIN_DISTANCE } from '../../constants'

export default class Line extends Base {
  constructor ({
    container,
    annotation,
    page,
    viewport,
    eventBus,
    layer,
    show = false
  }) {
    super({
      container,
      annotation,
      page,
      viewport,
      eventBus,
      show
    })

    this.layer = layer
    this.hidden = true
    this.initial = false
    this.outline = null

    this.start = null
    this.end = null

    this.newStart = null
    this.newEnd = null

    this.svgElement = null

    this.startCircle = null
    this.endCircle = null

    this.show = show

    this.onMousedown = this.handleMouseDown.bind(this)
    this.onMouseup = this.handleMouseUp.bind(this)
    this.onMousemove = this.handleMouseMove.bind(this)
    this.onDelete = this.handleDelete.bind(this)
    this.onKeydown = this.handleKeydown.bind(this)
    this.render()
    
  }

  render () {
    const annotation = this.annotation
    const { start, end } = this.getActualPoint(
      this.viewport,
      this.scale
    )
    this.start = start
    this.end = end
    // console.log(annotation)
    const uuid = uuidv4();
    const actualbdwidth = (this.annotation.lineWidth || this.annotation.width || 2) * this.scale
    const rect = this.calculate(start, end)

    this.svgElement = document.createElementNS("http://www.w3.org/2000/svg", "svg");

    this.svgElement.setAttribute("viewBox", `${rect.left} ${rect.top} ${rect.width} ${rect.height}`)
    let line = `<line
      x1=${start.x}
      y1=${start.y}
      x2=${end.x}
      y2=${end.y}
      stroke=${annotation.borderColor || annotation.color}
      stroke-width=${actualbdwidth}
      stroke-opacity=${annotation.transparency}`
  
    if (annotation.arrow) {
      const angle = Math.atan2(end.x + 10 - start.x, end.y + 10 - start.y);
  
      // 计算箭头的三个点的坐标
      const point1X = end.x - 10 * Math.cos(angle - Math.PI / 5);
      const point1Y = end.y - 10 * Math.sin(angle - Math.PI / 5);
      const point2X = end.x - 10 * Math.cos(angle + Math.PI / 5);
      const point2Y = end.y - 10 * Math.sin(angle + Math.PI / 5);
      line = `<defs>
        <marker
          id=${uuid}
          markerWidth=${actualbdwidth * 2}
          markerHeight=${actualbdwidth * 3}
          refX=3
          refY=2
          orient="auto"
          markerUnits="strokeWidth"
        >
          <polyline
            points="0.25,0.5 3.25,2 0.25,4"
            stroke=${annotation.borderColor}
            stroke-width=1
            fill="none"
            stroke-dasharray=100
            stroke-opacity=${annotation.transparency}
          />
        </marker>
      </defs>` + line + ` markerEnd=url(#${uuid})></line>`
    } else {
      line = line + `></line>`
    }
    this.svgElement.innerHTML = line

    this.annotationContainer = document.createElement('div')
    this.annotationContainer.append(this.svgElement)
    this.annotationContainer.id = annotation.name
    this.annotationContainer.className = 'annotation'
    this.annotationContainer.style.top = rect.top + 'px'
    this.annotationContainer.style.left = rect.left + 'px'
    this.annotationContainer.style.width = rect.width + 'px'
    this.annotationContainer.style.height = rect.height + 'px'
    this.svgElement.addEventListener('mousedown', this.handleClick.bind(this))

    this.container.append(this.annotationContainer)

    this.outerLineContainer = document.createElement('div')
    this.outerLineContainer.className = 'outline-container'
    this.deletetButton = createSvg(
      "svg",
      {
        width: "30",
        height: "30",
        viewBox: "0 0 30 30",
        fill: "none",
        class: 'delete-button'
      },
      {
        width: '30px',
        height: '30px'
      }
    );
    this.deletetButton.addEventListener('click', this.onDelete)
    this.deletetButton.innerHTML = this.deleteSvgStr
    
    const left = (rect.left + rect.width + MARGIN_DISTANCE -30) + 'px'
    const top = (rect.top - MARGIN_DISTANCE -30) + 'px'
    this.deletetButton.style.left = left
    this.deletetButton.style.top = top
    
    const outerLine = document.createElementNS("http://www.w3.org/2000/svg", "svg");
    
    outerLine.setAttribute("data-id", "move");
    outerLine.setAttribute("height", rect.height);
    outerLine.setAttribute("viewBox", `${rect.left} ${rect.top} ${rect.width + MARGIN_DISTANCE * 2} ${rect.height + MARGIN_DISTANCE * 2}`)
    outerLine.style.position = 'absolute'
    outerLine.style.zIndex = 1
    outerLine.style.left = `${rect.left - MARGIN_DISTANCE}px`
    outerLine.style.top = `${rect.top - MARGIN_DISTANCE}px`
    outerLine.style.width = `${rect.width + MARGIN_DISTANCE * 2}px`
    outerLine.style.height = `${rect.height + MARGIN_DISTANCE * 2}px`
    this.outerLine = outerLine
    
    this.startCircle = createSvg(
      "circle",
      {
        class: "start",
        'data-id': "start",
        fill: "#1460F3",
        cx: start.x + MARGIN_DISTANCE,
        cy: start.y + MARGIN_DISTANCE,
        r: 6
      }
    );
    this.endCircle = createSvg(
      "circle",
      {
        class: "end",
        'data-id': "end",
        fill: "#1460F3",
        cx: end.x + MARGIN_DISTANCE,
        cy: end.y + MARGIN_DISTANCE,
        r: 6
      }
    );
    this.moveLine = createSvg(
      "line",
      {
        class: 'move',
        'data-id': "move",
        x1: start.x + MARGIN_DISTANCE,
        y1: start.y + MARGIN_DISTANCE,
        x2: end.x + MARGIN_DISTANCE,
        y2: end.y + MARGIN_DISTANCE,
        fill: 'transparent',
        stroke: 'transparent',
        'stroke-width': 5
      }
    )
    this.outerLine.append(this.moveLine)
    this.outerLine.append(this.startCircle)
    this.outerLine.append(this.endCircle)
    this.outerLineContainer.append(this.outerLine)
    this.outerLineContainer.append(this.deletetButton)
    this.initial = true
  }

  handleClick () {
    if (!this.hidden) return
    this.hidden = false
    this.updateTool()
    onClickOutside([this.svgElement, this.outerLine, this.deletetButton], this.handleOutside.bind(this))
  }

  handleOutside () {
    this.hidden = !this.hidden
    if (this.layer.selectedElementName === this.annotation.name) {
      this.layer.selectedElementName = null
    }
    this.outerLine.removeEventListener('mousedown', this.onMousedown)
    this.outerLine.removeEventListener('touchstart', this.onMousedown)
    this.outerLineContainer.remove()
  }

  getActualPoint (
    viewport,
    s,
  ) {
    const annotation = this.annotation
    if (!annotation.linePoints && annotation.start && annotation.end) {
      annotation.linePoints = annotation.start + ',' + annotation.end
    }
    let [startX, startY, endX, endY] = annotation.linePoints.split(',')

    const { width, height, rotation  } = viewport;
    let x1, y1, x2, y2;
    if (rotation === 0) {
      x1 = startX * s;
      y1 = height - startY * s;
      x2 = endX * s;
      y2 = height - endY * s;
    } else if (rotation === 90) {
      x1 = startY * s;
      y1 = startX * s;
      x2 = endY * s;
      y2 = endX * s;
    } else if (rotation === 180) {
      x1 = width - startX * s;
      y1 = startY * s;
      x2 = width - endX * s;
      y2 = endY * s;
    } else {
      x1 = width - startY * s;
      y1 = height - startX * s;
      x2 = width - endY * s;
      y2 = height - endX * s;
    }
    return {
      start: {
        x: x1,
        y: y1,
      },
      end: {
        x: x2,
        y: y2,
      },
    };
  };

  getInitialPoint () {
    const s = this.scale
    const startX = this.start.x
    const startY = this.start.y
    const endX = this.end.x
    const endY = this.end.y

    const { width, height, rotation  } = this.viewport;
    let x1, y1, x2, y2;
    if (rotation === 0) {
      x1 = startX / s;
      y1 = (height - startY) / s; 
      x2 = endX / s;
      y2 = (height - endY) / s;
    } else if (rotation === 90) {
      x1 = startY / s;
      y1 = startX / s;
      x2 = endY / s;
      y2 = endX / s;
    } else if (rotation === 180) {
      x1 =( width - startX) / s;
      y1 = startY / s;
      x2 = (width - endX) / s;
      y2 = endY / s;
    } else {
      x1 = (height - startY) / s
      y1 = (width - startX) / s
      x2 = (height - endY) / s
      y2 = (width - endX)  / s
    }
    return {
      start: {
        x: x1,
        y: y1,
      },
      end: {
        x: x2,
        y: y2,
      },
    };
  }

  calculate (start, end) {
    const initRect = this.rectCalc({ start, end });
    const actualbdwidth = (this.annotation.width || this.annotation.lineWidth || 2) * this.scale
    if (this.annotation.arrow) {
      return {
        top: initRect.top - 5 * 2,
        left: initRect.left - 5 * 2,
        width: initRect.width + 5 * 4,
        height: initRect.height + 5 * 4,
      }
    } else {
      return {
        top: initRect.top - actualbdwidth / 2,
        left: initRect.left - actualbdwidth / 2,
        width: initRect.width + actualbdwidth,
        height: initRect.height + actualbdwidth,
      }
    }
  }

  rectCalc ({ start, end }) {
    return {
      top: start.y < end.y ? start.y : end.y,
      left: start.x < end.x ? start.x : end.x,
      width: Math.abs(end.x - start.x),
      height: Math.abs(end.y - start.y),
    };
  };

  updateTool () {
    if (!this.initial) return
    if (this.hidden) {
      if (this.layer.selectedElementName === this.annotation.name) {
        this.layer.selectedElementName = null
      }
      this.outerLineContainer.remove()
      document.removeEventListener('keydown', this.onKeydown)
    } else {
      if (this.layer.selectedElementName !== this.annotation.name) {
        this.layer.selectedElementName = this.annotation.name
      }
      this.container.append(this.outerLineContainer)
      this.outerLine.addEventListener('mousedown', this.onMousedown)
      this.outerLine.addEventListener('touchstart', this.onMousedown)
      document.addEventListener('keydown', this.onKeydown)
    }
  }

  handleMouseDown (event) {
    if (this.layer.tool) {
      event.stopPropagation()
    }
    const operatorId = event.target.getAttribute('data-id')
  
    const { pageX, pageY } = getClickPoint(event)
    this.startState = {
      operator: operatorId,
      clickX: pageX,
      clickY: pageY,
    }
    document.addEventListener('mousemove', this.onMousemove)
    document.addEventListener('mouseup', this.onMouseup)
    document.addEventListener('touchmove', this.onMousemove)
    document.addEventListener('touchend', this.onMouseup)
    
  };

  handleMouseMove (event) {
    if (event.type === 'touchmove') {
      document.body.style.overscrollBehavior = 'none';
    }
    this.moving = true
    const { pageX, pageY } = getClickPoint(event)
    const start = this.start
    const end = this.end
    const startState = this.startState

    const { width, height } = this.viewport
    if (startState.operator === 'start') {
      let x1 = pageX - (startState.clickX - this.start.x);
      let y1 = pageY - (startState.clickY - this.start.y);
      x1 = Math.min(width, x1)
      x1 = Math.max(0, x1)
      y1 = Math.min(height, y1)
      y1 = Math.max(0, y1)

      this.update({
        start: { x: x1, y: y1 },
        end,
      });
    } else if (startState.operator === 'end') {
      let x2 = pageX - (startState.clickX - this.end.x);
      let y2 = pageY - (startState.clickY - this.end.y);
      x2 = Math.min(width, x2)
      x2 = Math.max(0, x2)
      y2 = Math.min(height, y2)
      y2 = Math.max(0, y2)
      this.update({
        start,
        end: { x: x2, y: y2 },
      });
    } else if (startState.operator === 'move') {
      let x1 = pageX - (startState.clickX - this.start.x);
      let y1 = pageY - (startState.clickY - this.start.y);
      let x2 = pageX - (startState.clickX - this.end.x);
      let y2 = pageY - (startState.clickY - this.end.y);

      const rect = {
        width: Math.abs(start.x -end.x),
        height: Math.abs(start.y -end.y)
      }
      if (x1 < x2) {
        x1 = Math.max(0, x1)
        x1 = Math.min(width - rect.width, x1)
        x2 = Math.max(rect.width, x2)
        x2 = Math.min(width, x2)
      } else {
        x2 = Math.max(0, x2)
        x2 = Math.min(width - rect.width, x2)
        x1 = Math.max(rect.width, x1)
        x1 = Math.min(width, x1)
      }

      if (y1 < y2) {
        y1 = Math.max(0, y1)
        y1 = Math.min(height - rect.height, y1)
        y2 = Math.max(rect.height, y2)
        y2 = Math.min(height, y2)
      } else {
        y2 = Math.max(0, y2)
        y2 = Math.min(height - rect.height, y2)
        y1 = Math.max(rect.height, y1)
        y1 = Math.min(height, y1)
      }

      this.update({
        start: { x: x1, y: y1 },
        end: { x: x2, y: y2 },
      });
    }
  }

  handleDelete (event) {
    if (this.layer.tool) {
      event.stopPropagation()
    }
    this.handleOutside()
    this.annotationContainer.remove()
    this.annotation.isDelete = true
    this.eventBus.dispatch('annotationChange', {
      type: 'delete',
      annotation: {
        operate: "del-annot",
        name: this.annotation.name,
        page: this.page
      }
    })
  }

  handleKeydown (event) {
    if (event.key === 'Delete' || event.key === 'Backspace') {
      this.handleDelete()
    }
  }


  handleMouseUp (event) {
    if (this.layer.tool) {
      event.stopPropagation()
    }
    if (event.type === 'touchend') {
      document.body.style.overscrollBehavior = 'auto';
    }
    this.moving = false
    document.removeEventListener('mousemove', this.onMousemove)
    document.removeEventListener('mouseup', this.onMouseup)
    document.removeEventListener('touchmove', this.onMousemove)
    document.removeEventListener('touchend', this.onMouseup)
    
    const { pageX, pageY } = getClickPoint(event)
    if (pageX === this.startState.clickX && pageY === this.startState.clickY) return

    this.start = this.newStart
    this.end = this.newEnd

    const annotation = this.annotation
    const { start, end } = this.getInitialPoint()

    const linePoints = start.x + ',' + start.y + ',' + end.x + ',' + end.y
    annotation.linePoints = linePoints

    this.eventBus.dispatch('annotationChange', {
      type: 'modify',
      annotation: {
        operate: "mod-annot",
        name: annotation.name,
        page: this.page,
        linePoints: linePoints,
      }
    })
  }

  update ({ start, end }) {
    const rect = this.calculate(start, end)

    this.newStart = start
    this.newEnd = end

    this.svgElement.setAttribute("width", rect.width);
    this.svgElement.setAttribute("height", rect.height);
    this.svgElement.setAttribute("viewBox", `${rect.left} ${rect.top} ${rect.width} ${rect.height}`)

    const line = this.svgElement.querySelector('line')
    line.setAttribute("x1", start.x)
    line.setAttribute("y1", start.y)
    line.setAttribute("x2", end.x)
    line.setAttribute("y2", end.y)

    this.annotationContainer.style.top = rect.top + 'px'
    this.annotationContainer.style.left = rect.left + 'px'
    this.annotationContainer.style.width = rect.width + 'px'
    this.annotationContainer.style.height = rect.height + 'px'

    this.outerLine.setAttribute("height", rect.height);
    this.outerLine.setAttribute("viewBox", `${rect.left} ${rect.top} ${rect.width + MARGIN_DISTANCE * 2} ${rect.height + MARGIN_DISTANCE * 2}`)
    this.outerLine.style.left = `${rect.left - MARGIN_DISTANCE}px`
    this.outerLine.style.top = `${rect.top - MARGIN_DISTANCE}px`
    this.outerLine.style.width = `${rect.width + MARGIN_DISTANCE * 2}px`
    this.outerLine.style.height = `${rect.height + MARGIN_DISTANCE * 2}px`
    const left = (rect.left + rect.width + MARGIN_DISTANCE -30) + 'px'
    const top = (rect.top - MARGIN_DISTANCE -30) + 'px'
    this.deletetButton.style.left = left
    this.deletetButton.style.top = top
    
    this.startCircle.setAttribute("cx", start.x + MARGIN_DISTANCE)
    this.startCircle.setAttribute("cy", start.y + MARGIN_DISTANCE)
    this.endCircle.setAttribute("cx", end.x + MARGIN_DISTANCE)
    this.endCircle.setAttribute("cy", end.y + MARGIN_DISTANCE)
    
    this.moveLine.setAttribute("x1", start.x + MARGIN_DISTANCE)
    this.moveLine.setAttribute("y1", start.y + MARGIN_DISTANCE)
    this.moveLine.setAttribute("x2", end.x + MARGIN_DISTANCE)
    this.moveLine.setAttribute("y2", end.y + MARGIN_DISTANCE)
  }
}
