import { defineStore } from 'pinia'

export const useViewerStore = defineStore({
  id: 'viewer',
  state: () => ({
    fullMode: false,
    currentPage: 0,
    searchStatus: false,
    scale: '',
    themeMode: 'Light',
    pageMode: 0,
    scrollMode: 'Vertical',
    pageLabels: [],
    zoomLevel: ['auto', 'page-fit', 0.5, 1, 1.25, 1.5, 2, 2.5],
    activeTab: 0,
    activeLeftPanel: 'THUMBS',
    activeLeftPanelTab: 'THUMBS',
    activeStickNote: false,
    activeActiveMeasure: false,
    activeHand: false,
    openElements: {
      header: true,
      leftPanel: false
    },
    headers: [
      {
        type: 'toggleElementButton',
        img: 'icon-header-sidebar-line',
        element: 'leftPanel',
        dataElement: 'leftPanelButton',
        title: 'Left Panel'
      },
      {
        type: 'divider',
        hidden: [
          'small-mobile'
        ]
      },
      {
        type: 'pageDisplayButton',
        dataElement: 'pageDisplayButton',
        element: 'pageDisplayButton',
        title: 'Page Display'
      },
      {
        type: 'fullScreenButton',
        dataElement: 'fullScreenButton',
        element: 'fullScreenButton',
        title: 'Full Screen'
      },
      {
        type: 'handToolButton',
        dataElement: 'handToolButton',
        element: 'handToolButton',
        title: 'Pan Tool'
      },
      {
        type: 'divider',
        hidden: [
          'small-mobile'
        ]
      },
      {
        type: 'zoomOverlay',
        dataElement: 'zoomOverlayButton',
        element: 'zoomOverlay',
        hiddenOnMobileDevice: true
      },
      {
        type: 'divider',
        hidden: [
          'small-mobile'
        ]
      },
      {
        type: 'themeMode',
        dataElement: 'themeMode',
        element: 'themeMode'
      },
      {
        type: 'stickyNoteButton',
        dataElement: 'stickyNoteButton',
        element: 'stickyNoteButton',
        hidden: false,
      },
      {
        type: 'markup',
        dataElement: 'markup',
        element: 'markup',
        hidden: false,
      },
      {
        type: 'measureButton',
        dataElement: 'measureButton',
        element: 'measureButton',
        hidden: true,
      },
      {
        type: 'compareButton',
        dataElement: 'compareButton',
        element: 'compareButton',
        hidden: true,
      },
      {
        type: 'saveButton',
        dataElement: 'saveButton',
        element: 'saveButton',
        hidden: true,
      }
    ],
    rightHeaders: [
      {
        type: 'openFileButton',
        dataElement: 'openFileButton',
        element: 'openFileButton',
        title: 'Open File'
      },
      {
        type: 'searchButton',
        dataElement: 'searchButton',
        element: 'searchButton',
        title: 'Search'
      },
      {
        type: 'downloadButton',
        dataElement: 'downloadButton',
        element: 'downloadButton',
        title: 'Download'
      },
      {
        type: 'printButton',
        dataElement: 'printButton',
        element: 'printButton',
        title: 'Print'
      }
    ],
    disabledElements: null
  }),
  getters: {
    getFullMode () {
      return this.fullMode
    },
    getCurrentPage () {
      return this.currentPage
    },
    getThemeMode () {
      return this.themeMode
    },
    getPageMode () {
      return this.pageMode
    },
    getScrollMode () {
      return this.scrollMode
    },
    getActiveHeaderItems () {
      return this.headers
    },
    getActiveRightHeaderItems () {
      return this.rightHeaders
    },
    isLeftPanelOpen: (state) => {
      return (dataElement) => state.openElements[dataElement] && !state.disabledElements?.[dataElement]?.disabled
    },
    getActiveLeftPanel () {
      return this.activeLeftPanel
    },
    getActiveLeftPanelTab () {
      return this.activeLeftPanelTab
    },
    getScale () {
      return Math.round(this.scale * 100)
    },
    getActiveStickNote () {
      return this.activeStickNote
    },
    getActiveMeasure () {
      return this.activeActiveMeasure
    },
    getActiveHand () {
      return this.activeHand
    },
    getSearchStatus () {
      return this.searchStatus
    }
  },
  actions: {
    resetSetting () {
      this.fullMode = false
      this.currentPage = 0
      this.scale = '',
      this.themeMode = 'Light'
      this.pageMode = 0
      this.scrollMode = 'Vertical'
      this.activeTab = 0
      this.activeLeftPanel = 'THUMBS',
      this.activeLeftPanelTab = 'THUMBS'
      this.searchStatus = false
      this.openElements = {
        header: true,
        leftPanel: false
      }
    },
    setFullMode (fullMode) {
      return this.fullMode = fullMode
    },
    setCurrentPage (currentPage) {
      this.currentPage = currentPage
    },
    setThemeMode (themeMode) {
      this.themeMode = themeMode
    },
    setPageMode (mode) {
      this.pageMode = mode
    },
    setScrollMode (mode) {
      this.scrollMode = mode
    },
    setCurrentScale (scale) {
      this.scale = scale
    },
    openElement (dataElement) {
      this.openElements[dataElement] = true
    },
    closeElement (dataElement) {
      this.openElements[dataElement] = false
    },
    toggleElement (dataElement) {
      if (this.disabledElements?.[dataElement]?.disabled) {
        return
      }
      if (this.openElements[dataElement]) {
        this.closeElement(dataElement)
      } else {
        this.openElement(dataElement)
      }
    },
    setActiveLeftPanel (dataElement) {
      this.activeLeftPanel = dataElement
    },
    setActiveLeftPanelTab (dataElement) {
      this.activeLeftPanelTab = dataElement
    },
    toggleActiveStickNote () {
      this.activeStickNote = !this.activeStickNote
    },
    closeActiveStickNote () {
      this.activeStickNote = false
    },
    toggleActiveMeasure () {
      this.activeActiveMeasure = !this.activeActiveMeasure
    },
    closeActiveMeasure () {
      this.activeActiveMeasure = false
    },
    toggleActiveHand (value) {
      this.activeHand = value
    },
    closeActiveHand () {
      this.activeHand = false
    },
    setSearchStatus (value) {
      this.searchStatus = value
    }
  }
})
